# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


from ._binary import BinaryParser, BinaryWriter
import collections
import zx


class UnifiedSnapshot(zx.MachineSnapshot):
    pass


_V1_FORMAT = '1.45'
_V2_FORMAT = '2.x'
_V3_FORMAT = '3.x'

def _get_format_version(fields):
    if 'ticks_count_low' in fields:
        return _V3_FORMAT

    if 'pc2' in fields:
        return _V2_FORMAT

    return _V1_FORMAT


class Z80Snapshot(zx.MachineSnapshot):
    _MEMORY_PAGE_ADDRS = {4: 0x8000, 5: 0xc000, 8: 0x4000}

    def get_unified_snapshot(self):
        # Bit 7 of the stored R value is not significant and
        # shall be taken from bit 0 of flags1.
        flags1 = self['flags1']
        r = (self['r'] & 0x7f) | ((flags1 & 0x1) << 7)

        flags2 = self['flags2']
        int_mode = flags2 & 0x3
        if int_mode not in [0, 1, 2]:
            raise zx.Error('Invalid interrupt mode %d.' % int_mode)

        format_version = _get_format_version(self._fields)

        pc = self['pc'] if format_version == _V1_FORMAT else self['pc2']

        processor_fields = {
            'bc': self['bc'],
            'de': self['de'],
            'hl': self['hl'],
            'af': zx.make16(hi=self['a'], lo=self['f']),
            'ix': self['ix'],
            'iy': self['iy'],
            'alt_bc': self['alt_bc'],
            'alt_de': self['alt_de'],
            'alt_hl': self['alt_hl'],
            'alt_af': zx.make16(hi=self['alt_a'], lo=self['alt_f']),
            'pc': pc,
            'sp': self['sp'],
            'ir': zx.make16(hi=self['i'], lo=r),
            'iff1': 0 if self['iff1'] == 0 else 1,
            'iff2': 0 if self['iff2'] == 0 else 1,
            'int_mode': int_mode }

        ticks_per_frame = 69888  # TODO
        quarter_tstates = ticks_per_frame // 4

        fields = {
            'processor_snapshot': zx.ProcessorSnapshot(processor_fields),
            'border_color': (flags1 >> 1) & 0x7,

            # Give the snapshot a chance to execute at least one
            # instruction without firing up an interrupt.
            'ticks_since_int': ticks_per_frame - 23,
        }

        if 'ticks_count_high' in self:
            ticks_high = self['ticks_count_high']
            ticks_low = self['ticks_count_low']
            ticks_since_int = (((ticks_high + 1) % 4 + 1) * quarter_tstates -
                                   (ticks_low + 1))
            fields['ticks_since_int'] = ticks_since_int

        # Determine machine kind.
        # TODO: Not used currently?
        if format_version == _V1_FORMAT:
            machine_kind = 'ZX Spectrum 48K'
        else:
            hardware_mode = self['hardware_mode']
            flags3 = self['flags3']
            flags3_bit7 = (flags3 & 0x80) >> 7
            if hardware_mode == 0 and not flags3_bit7:
                machine_kind = 'ZX Spectrum 48K'
            else:
                raise zx.Error('Unsupported type of emulated machine.',
                               id='unsupported_machine')

        # Handle memory blocks.
        memory_blocks = fields.setdefault('memory_blocks', [])
        if 'memory_snapshot' in self:
            memory_blocks.append((0x4000, self['memory_snapshot']))
        else:
            assert machine_kind == 'ZX Spectrum 48K', machine_kind  # TODO
            for block in self['memory_blocks']:
                page_no = block['page_no']
                image = block['image']
                memory_blocks.append((self._MEMORY_PAGE_ADDRS[page_no], image))

        return UnifiedSnapshot(fields)


class Z80SnapshotsFormat(zx.SnapshotsFormat):
    _PRIMARY_HEADER = [
        'B:a', 'B:f', '<H:bc', '<H:hl', '<H:pc', '<H:sp', 'B:i', 'B:r',
        'B:flags1', '<H:de', '<H:alt_bc', '<H:alt_de', '<H:alt_hl',
        'B:alt_a', 'B:alt_f', '<H:iy', '<H:ix', 'B:iff1', 'B:iff2', 'B:flags2']

    _EXTRA_HEADERS_SIZE_FIELD = [
        '<H:extra_headers_size']

    _EXTRA_HEADER = [
        '<H:pc2', 'B:hardware_mode', 'B:misc1', 'B:misc2', 'B:flags3',
        'B:port_fffd_value', '16B:sound_chip_registers']

    _EXTRA_HEADER2 = [
        '<H:ticks_count_low', 'B:ticks_count_high', 'B:spectator_flag',
        'B:mgt_rom_paged', 'B:multiface_rom_paged',
        'B:memory_at_0000_1fff_is_ram', 'B:memory_at_2000_3fff_is_ram',
        '10B:keyboard_mappings', '10B:keyboard_mappings_keys',
        'B:mgt_type', 'B:disciple_inhibit_button_status',
        'B:disciple_inhibit_flag']

    _EXTRA_HEADER2b = [
        'B:last_write_to_port_1ffd']

    _MEMORY_BLOCK_HEADER = [
        '<H:compressed_size', 'B:page_no']

    _RAW_MEMORY_BLOCK_SIZE_VALUE = 0xffff

    def _uncompress(self, compressed_image, uncompressed_size):
        MARKER = 0xed
        input = list(compressed_image)
        output = []
        while input:
            if len(input) >= 4 and input[0] == MARKER and input[1] == MARKER:
                count = input[2]
                filler = input[3]
                output.extend([filler] * count)
                del input[0:4]
            else:
                output.append(input.pop(0))

        if len(output) != uncompressed_size:
            raise zx.Error('Corrupted compressed memory block.')

        return bytes(output)

    def _parse_memory_block(self, parser):
        BLOCK_SIZE = 16 * 1024
        fields = parser.parse(self._MEMORY_BLOCK_HEADER)
        compressed_size = fields['compressed_size']
        if compressed_size == self._RAW_MEMORY_BLOCK_SIZE_VALUE:
            raw_image = parser.extract_block(BLOCK_SIZE)
        else:
            compressed_image = parser.extract_block(compressed_size)
            raw_image = self._uncompress(compressed_image, BLOCK_SIZE)
        return collections.OrderedDict(page_no=fields['page_no'],
                                       image=raw_image)

    def parse(self, image):
        # Parse headers.
        parser = BinaryParser(image)
        fields = collections.OrderedDict(id='z80_snapshot')
        fields.update(parser.parse(self._PRIMARY_HEADER))

        if fields['pc'] == 0:
            fields.update(parser.parse(self._EXTRA_HEADERS_SIZE_FIELD))

        if 'extra_headers_size' in fields:
            extra_headers = parser.extract_block(fields['extra_headers_size'])
            extra_parser = BinaryParser(extra_headers)
            fields.update(extra_parser.parse(self._EXTRA_HEADER))

            if extra_parser:
                fields.update(extra_parser.parse(self._EXTRA_HEADER2))

            if extra_parser:
                fields.update(extra_parser.parse(self._EXTRA_HEADER2b))

            if extra_parser:
                raise zx.Error('Too many headers in Z80 snapshot.',
                               id='too_many_z80_snapshot_headers')

        # Parse memory snapshot.
        if _get_format_version(fields) == _V1_FORMAT:
            compressed = (fields['flags1'] & 0x20) != 0
            if not compressed:
                if parser.get_rest_size() != 48 * 1024:
                    raise zx.Error('The snapshot is too large.')
                image = parser.extract_rest()
            else:
                assert 0  # TODO: Support compressed v1 memory snapshots.

            fields['memory_snapshot'] = image
        else:
            memory_blocks = fields.setdefault('memory_blocks', [])
            while parser:
                block = self._parse_memory_block(parser)
                memory_blocks.append(block)

        return Z80Snapshot(fields)

    # TODO: Rework to generate an internal representation of the
    #       format and then generate its binary version.
    def make(self, state):
        # TODO: The z80 format cannot represent processor states in
        #       the middle of IX- and IY-prefixed instructions, so
        #       such situations need some additional processing.
        # TODO: Check for similar problems with other state attributes.
        assert state.get_iregp_kind() == 'hl'

        flags1 = 0
        flags2 = 0

        # Bit 7 of the stored R value is not signigicant and
        # shall be taken from bit 0 of flags1.
        r = state.get_r_reg()
        flags1 |= (r & 0x80) >> 7
        r &= 0x7f

        border_color = state.get_border_color()
        assert 0 <= border_color <= 7
        flags1 |= border_color << 1

        int_mode = state.get_int_mode()
        assert int_mode in [0, 1, 2]  # TODO
        flags2 |= int_mode

        # Write v1 header.
        # TODO: Support other versions.
        writer = BinaryWriter()
        writer.write(
            self._PRIMARY_HEADER,
            a=state.get_a(), f=state.get_f(), bc=state.get_bc(), hl=state.get_hl(),
            pc=state.get_pc(), sp=state.get_sp(), i=state.get_i(), r=r,
            flags1=flags1, de=state.get_de(),
            alt_bc=state.get_alt_bc(), alt_de=state.get_alt_de(),
            alt_hl=state.get_alt_hl(),
            alt_a=state.get_alt_a(), alt_f=state.get_alt_f(),
            iy=state.get_iy(), ix=state.get_ix(),
            iff1=state.get_iff1(), iff2=state.get_iff2(), flags2=flags2)

        # Write memory snapshot.
        writer.write_block(state.get_memory_block(0x4000, size=48 * 1024))

        return writer.get_image()



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


import struct, zx


class BinaryParser(object):
    def __init__(self, image):
        self.image = image
        self.pos = 0

    def get_rest_size(self):
        return len(self.image) - self.pos

    def is_eof(self):
        return self.get_rest_size() == 0

    def __bool__(self):
        return not self.is_eof()

    def extract_block(self, size):
        if size > self.get_rest_size():
            raise zx.Error('Binary image is too short.',
                           id='binary_image_too_short')

        begin = self.pos
        self.pos += size
        return self.image[begin:self.pos]

    def extract_rest(self):
        return self.extract_block(len(self.image) - self.pos)

    def parse_field(self, format, id):
        size = struct.calcsize(format)
        value = struct.unpack(format, self.extract_block(size))
        if len(value) == 1:
            value = value[0]
        return value

    def parse(self, format):
        res = dict()
        for field in format:
            if isinstance(field, str):
                field_format, field_id = field.split(':', maxsplit=1)
            else:
                # TODO: Remove this branch once all tuple formats
                # are eliminated.
                field_id, field_format = field
            res[field_id] = self.parse_field(field_format, field_id)
        return res


class BinaryWriter(object):
    def __init__(self):
        self._chunks = []

    def write_block(self, block):
        self._chunks.append(block)

    def write(self, format, **values):
        for field in format:
            field_format, field_id = field.split(':', maxsplit=1)
            self.write_block(struct.pack(field_format, values[field_id]))

    def get_image(self):
        return b''.join(self._chunks)



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


from ._data import Data, File, FileFormat, SoundFile, SoundFileFormat
from ._keyboard import KEYS_INFO
from ._machine import Spectrum48
from ._main import main, ProcessorSnapshot, MachineSnapshot, SnapshotsFormat
from ._rom import get_rom_image
from ._rzx import parse_rzx, make_rzx
from ._tap import TAPFileFormat
from ._tzx import TZXFileFormat
from ._utils import make16
from ._wav import WAVFileFormat
from ._z80snapshot import Z80SnapshotsFormat


class Error(Exception):
    """Basic exception for the whole ZX module."""
    def __init__(self, reason, id=None):
        super().__init__(reason)

        if id:
            self.id = id



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


from ._binary import BinaryParser
from ._tape import *
import zx


class TAPFile(zx.SoundFile):
    _TICKS_FREQ = 3500000

    def __init__(self, fields):
        zx.SoundFile.__init__(self, TAPFileFormat, fields)

    def get_pulses(self):
        level = False
        blocks = self['blocks']
        for data in blocks:
            # The block itself.
            for pulse in get_standard_block_pulses(data):
                yield (level, pulse)
                level = not level

            # Pause. Skip, if it's the last block.
            if data is not blocks[-1]:
                yield (level, self._TICKS_FREQ)  # 1s.


class TAPFileFormat(zx.SoundFileFormat):
    _NAME = 'TAP'

    def _parse_block(self, parser):
        size = parser.parse_field('<H', 'block_size')
        return parser.extract_block(size)

    def parse(self, image):
        parser = BinaryParser(image)

        # Parse blocks.
        blocks = []
        while not parser.is_eof():
            blocks.append(self._parse_block(parser))

        return TAPFile({'blocks': blocks})



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


import gc, unittest


class test_spectrum48_rom(unittest.TestCase):
    def runTest(self):
        import zx
        rom = zx.get_rom_image('ZX Spectrum 48K')
        assert type(rom) is bytes
        assert len(rom) == 0x4000
        assert rom.startswith(b'\xf3\xaf\x11\xff')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


import sys


class Reg(object):
    def __init__(self, name):
        self.name = name


class RegPair(object):
    def __init__(self, name):
        self.name = name


A = Reg('a')
B = Reg('b')
C = Reg('c')
D = Reg('d')
E = Reg('e')
H = Reg('h')
L = Reg('l')
R = Reg('r')

BC = RegPair('bc')
DE = RegPair('de')
HL = RegPair('hl')


class Command:
    pass


class Load(Command):
    def __init__(self, reg, value):
        self.reg = reg
        self.value = value


class OutA(Command):
    def __init__(self, x, y):
        self.beam_pos = x, y


class WriteAtHL(Command):
    def __init__(self, reg):
        self.reg = reg


class WriteScreenAtHL(Command):
    def __init__(self, x, y, reg):
        self.beam_pos = x, y
        self.reg = reg


class drawing_generator(object):
    SCREEN_WIDTH = 256
    SCREEN_HEIGHT = 192
    TICKS_PER_LINE = 224
    PIXELS_PER_TICK = 2

    def __init__(self, is_late_timings=False):
        self.is_late_timings = is_late_timings
        self._lines = []
        self._label = 0
        self._tick = 38

    def add_line(self, line):
        # print(line)
        self._lines.append(line)

    def add_label(self):
        n = self._label
        self._label += 1
        return 'l%d' % n

    def add_instr(self, instr, ticks):
        if type(ticks) is not list:
            ticks = [ticks]
        self.add_line('    %-24s; %5d  %5s' % (
            instr, self._tick, ' + '.join('%d' % x for x in ticks)))
        self._tick += sum(ticks)

    def generate_load(self, load):
        if isinstance(load.reg, RegPair):
            self.add_instr('ld %s, 0x%04x' % (load.reg.name, load.value), 10)
        else:
            self.add_instr('ld %s, 0x%02x' % (load.reg.name, load.value), 7)

    def generate_out_a(self, out):
        self.move_to_beam_pos(out.beam_pos, ticks_to_reserve=7)
        self.add_instr('out (0xfe), a', 11)

    def generate_write_at_hl(self, write):
        delay = self.get_memory_contention_delay(self._tick + 4)
        self.add_instr('ld (hl), %s' % write.reg.name, [7, delay])

    def generate_write_screen_at_hl(self, write):
        self.move_to_beam_pos(write.beam_pos, ticks_to_reserve=4)
        self.generate_write_at_hl(write)

    _COMMAND_GENERATORS = {
        Load: generate_load,
        OutA: generate_out_a,
        WriteAtHL: generate_write_at_hl,
        WriteScreenAtHL: generate_write_screen_at_hl,
    }

    def generate_command(self, command):
        self._COMMAND_GENERATORS[type(command)](self, command)

    def _get_reg_pair_clobbers(self, clobbers):
        if B in clobbers and C in clobbers:
            yield BC

        if D in clobbers and E in clobbers:
            yield DE

        if H in clobbers and L in clobbers:
            yield HL

    # TODO: Test and optimize.
    def generate_delay(self, delay, clobbers=[], step=0):
        while delay:
            # print('delay:', delay)
            if (delay < 4 or delay == 5) and step:
                delay += step
                continue

            if delay == 4:
                self.add_instr('nop', 4)
                break

            if delay == 6:
                done = False
                for rp in self._get_reg_pair_clobbers(clobbers):
                    self.add_instr('inc %s' % rp.name, 6)
                    done = True
                    break

                if done:
                    break

                if step:
                    delay += step
                    continue

                assert 0, ('No register pair to clobber to generate a '
                           'delay of %d ticks!' % delay)

            if delay == 7:
                self.add_instr('or 0', 7)
                break

            if delay == 9:
                # TODO: Other options are:
                #  ld a, r
                #  ld a, i
                #  ld i, a
                if R in clobbers:
                    self.add_instr('ld r, a', 9)
                    break
                elif step:
                    delay += step
                    continue
                else:
                    assert 0, ('Need to clobber the register R to generate a '
                               'delay of %d ticks!' % delay)

            if delay == 10:
                label = self.add_label();
                self.add_instr('jp %s' % label, 10)
                self.add_line(label + ':')
                break

            if delay == 13:
                # TODO: 13 = 4 + 9, where the 9 needs a clobber.
                assert 0

            NON_CLOBBERING_PATTERNS = {
                # TODO: Add the missing values.
                8: [4, 4],
                11: [4, 7],
                12: [4, 4, 4],
                14: [7, 7],
                15: [4, 4, 7],
                16: [4, 4, 4, 4],
                17: [7, 10],
                18: [4, 7, 7],
                19: [4, 4, 4, 7],
                20: [4, 4, 4, 4, 4],
                21: [7, 7, 7],
                22: [4, 4, 7, 7],
                23: [4, 4, 4, 4, 7],
                24: [4, 4, 4, 4, 4, 4],
                25: [4, 7, 7, 7],
                26: [4, 4, 4, 7, 7],
                27: [4, 4, 4, 4, 4, 7],
                28: [4, 4, 4, 4, 4, 4, 4],
                29: [4, 4, 7, 7, 7],
                30: [4, 4, 4, 4, 7, 7],
                31: [4, 4, 4, 4, 4, 4, 7],
            }

            if delay in NON_CLOBBERING_PATTERNS:
                pattern = NON_CLOBBERING_PATTERNS[delay]
                assert sum(pattern) == delay
                for d in pattern:
                    self.generate_delay(d)
                break

            # TODO: Use djnz loops when they are a good fit.
            LOOP_THRESHOLD = 2 * 16 + 7 - 5
            if delay >= LOOP_THRESHOLD:
                assert clobbers, 'No clobber register for a delay loop!'
                clobber = clobbers[0]

                n = (delay - 7 + 5) // 16 - 1
                n = min(n, 0xff)  # TODO: Support longer loops.
                assert n > 0

                self.add_instr('ld %s, %d' % (clobber.name, n), 7)
                delay -= 7

                label = self.add_label()
                ticks = n * 16 - 5
                self.add_line('%-28s; %5d  %5d' % (
                    label + ':', self._tick, ticks))
                self.add_line('    %-24s;        %5d' % (
                    'dec %s' % clobber.name, 4))
                self.add_line('    %-24s;        %5s' % (
                    'jr nz, %s' % label, '7 + 5'))

                self._tick += ticks
                delay -= ticks

                # Make sure the remaining delay won't cause us
                # trouble.
                assert (delay >= LOOP_THRESHOLD or
                        delay in NON_CLOBBERING_PATTERNS), delay
                continue

            assert 0, "Don't know how to generate a delay of %d ticks!" % delay

    def move_to_tick(self, tick, clobbers=[], step=0):
        assert tick >= self._tick
        self.generate_delay(tick - self._tick, clobbers, step)

    def align_tick(self, div, rem):
        aligned_tick = (self._tick + (div - rem - 1)) // div * div + rem
        self.move_to_tick(aligned_tick, clobbers=[A, B, C, R], step=div)

    def align_end_tick(self):
        self.align_tick(4, 1)

    def move_to_beam_pos(self, pos, ticks_to_reserve):
        BASE_TICK = 64 * self.TICKS_PER_LINE + 8 // 2
        if self.is_late_timings:
            BASE_TICK += 1
        x, y = pos
        target_tick = (BASE_TICK + y * self.TICKS_PER_LINE + x // 2 -
                       ticks_to_reserve)
        self.move_to_tick(target_tick, clobbers=[B])

    def get_memory_contention_delay(self, tick):
        # TODO: We sample ~INT during the last tick of the
        # previous instruction, so we add 1 to the contention
        # base to compensate that.
        CONT_BASE = 14335 + 1
        if self.is_late_timings:
            CONT_BASE += 1
        if tick < CONT_BASE:
            return 0

        if tick >= CONT_BASE + self.SCREEN_HEIGHT * self.TICKS_PER_LINE:
            return 0

        ticks_since_new_line = (tick - CONT_BASE) % self.TICKS_PER_LINE
        if ticks_since_new_line >= self.SCREEN_WIDTH / self.PIXELS_PER_TICK:
            return 0

        ticks_since_new_ula_cycle = ticks_since_new_line % 8
        delay = (0 if ticks_since_new_ula_cycle == 7 else
                 6 - ticks_since_new_ula_cycle)

        return delay

    def generate(self, *commands):
        for command in commands:
            self.generate_command(command)

        self.align_end_tick()

        self.add_line('    %-24s; %5d' % ('', self._tick))

    def emit_source(self):
        for line in self._lines:
            print(line)


is_late_timings = len(sys.argv) > 1 and sys.argv[1] == 'late_timings'

g = drawing_generator(is_late_timings)
g.generate(
    # Let the border be (mostly) yellow.
    Load(A, 6), OutA(0, -60),

    # Use the spare time to prepare the screen and attributes areas.
    Load(A, 0x00),
    Load(HL, 0x4000), WriteAtHL(A),
    Load(A, 0xff),
    Load(HL, 0x4001), WriteAtHL(A),
    Load(HL, 0x4002), WriteAtHL(A),

    Load(HL, 0x4100), WriteAtHL(A),
    Load(HL, 0x4101), WriteAtHL(A),
    Load(HL, 0x4102), WriteAtHL(A),

    Load(HL, 0x4200), WriteAtHL(A),
    Load(A, 0x00),
    Load(HL, 0x4201), WriteAtHL(A),
    Load(A, 0xff),
    Load(HL, 0x4202), WriteAtHL(A),

    Load(HL, 0x4300), WriteAtHL(A),
    Load(HL, 0x4301), WriteAtHL(A),
    Load(HL, 0x4302), WriteAtHL(A),

    Load(HL, 0x4400), WriteAtHL(A),
    Load(HL, 0x4401), WriteAtHL(A),
    Load(A, 0x00),
    Load(HL, 0x4402), WriteAtHL(A),

    Load(A, 0xff),
    Load(HL, 0x4500), WriteAtHL(A),
    Load(HL, 0x4501), WriteAtHL(A),
    Load(HL, 0x4502), WriteAtHL(A),

    Load(A, 0xff),
    Load(HL, 0x5820), WriteAtHL(A),
    Load(HL, 0x5841), WriteAtHL(A),
    Load(HL, 0x5862), WriteAtHL(A),

    # Draw eight colour lines, each line starting one tick later
    # and via that make the moment when the border value is
    # latched be visible.
    Load(A, 0), OutA(-16, -16),
    Load(A, 5), OutA(-14, -14),
    Load(A, 2), OutA(-12, -12),
    Load(A, 4), OutA(-10, -10),

    # This black line starts one chunk later.
    Load(A, 0), OutA(-8, -8),
    Load(A, 5), OutA(-6, -6),
    Load(A, 2), OutA(-4, -4),
    Load(A, 4), OutA(-2, -2),

    # Continue the frame with yellow border again.
    Load(A, 0), OutA(256 - 2, -1),
    Load(A, 6), OutA(256 + 64, -1),

    # This write is early enough to clear the chunk of pixels
    # before it is latched.
    Load(A, 0xff),
    Load(HL, 0x4000), WriteScreenAtHL(-10, 0, A),

    # But this one is too late.
    Load(A, 0x00),
    Load(HL, 0x4100), WriteScreenAtHL(-8, 1, A),

    # Similarly, for the second chunk in line, this is early
    # enough to clear it.
    Load(A, 0xff),
    Load(HL, 0x4201), WriteScreenAtHL(-10, 2, A),

    # But this is again too late. Meaning both the adjacent
    # chunks are latched during the same ULA delay.
    Load(A, 0x00),
    Load(HL, 0x4301), WriteScreenAtHL(-8, 3, A),

    # Now let's see when the third chunk is latched so we know
    # the length of the 16-pixel cycles.
    Load(A, 0xff),
    Load(HL, 0x4402), WriteScreenAtHL(6, 4, A),

    Load(A, 0x00),
    Load(HL, 0x4502), WriteScreenAtHL(8, 5, A),

    # Now write some attribute bytes. This write is early enough
    # to colour the attribute square with black before it's
    # latched the first time.
    Load(A, 0x00),
    Load(HL, 0x5820), WriteScreenAtHL(-10, 8, A),

    # This write is too late, so it should remain invisible.
    Load(A, 0x6d),
    WriteScreenAtHL(-8, 12, A),

    # Make sure the rest of the square is black.
    Load(A, 0x00),
    WriteScreenAtHL(128, 12, A),

    # Do the same with the 2nd attribute byte in line.
    Load(A, 0x00),
    Load(HL, 0x5841), WriteScreenAtHL(-10, 16, A),

    Load(A, 0x6d),
    WriteScreenAtHL(-8, 20, A),

    Load(A, 0x00),
    WriteScreenAtHL(128, 20, A),

    # Do it once again for the 3rd attribute byte in line.
    Load(A, 0x00),
    Load(HL, 0x5862), WriteScreenAtHL(6, 24, A),

    Load(A, 0x6d),
    WriteScreenAtHL(8, 28, A),

    Load(A, 0x00),
    WriteScreenAtHL(128, 28, A),
)
g.emit_source()



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


MASK16 = 0xffff


def make16(hi, lo):
    return ((hi << 8) | lo) & MASK16



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


import os


def load_image(filename):
    current_dir = os.path.dirname(__file__)
    filename = os.path.join(current_dir, 'roms', filename)
    with open(filename, mode='rb') as f:
        return f.read()


def get_rom_image(machine_kind):
    rom_filenames = {'ZX Spectrum 48K': 'Spectrum48.rom'}
    return load_image(rom_filenames[machine_kind])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


import cairo, gi, os, sys, time, zx
import collections
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk



# TODO: Move to the z80 project.
class ProcessorSnapshot(zx.Data):
    pass


class MachineSnapshot(zx.Data):
    pass


class SnapshotsFormat(zx.FileFormat):
    pass


class RZXFile(zx.Data):
    def __init__(self, recording):
        self._recording = recording


class RZXFileFormat(zx.FileFormat):
    def parse(self, image):
        recording = zx.parse_rzx(image)
        return RZXFile(recording)


def detect_file_format(image, filename_extension):
    KNOWN_FORMATS = [
        ('.rzx', b'RZX!', RZXFileFormat),
        ('.tap', None, zx.TAPFileFormat),
        ('.tzx', b'ZXTape!', zx.TZXFileFormat),
        ('.wav', b'RIFF', zx.WAVFileFormat),
        ('.z80', None, zx.Z80SnapshotsFormat),
    ]

    filename_extension = filename_extension.lower()

    # First, try formats without signatures.
    for ext, signature, format in KNOWN_FORMATS:
        if not signature and filename_extension == ext:
            return format

    # Then, look at the signature.
    if image:
        for ext, signature, format in KNOWN_FORMATS:
            if signature and image[:len(signature)] == signature:
                return format

    # Finally, just try to guess by the given extension.
    for ext, signature, format in KNOWN_FORMATS:
        if filename_extension == ext:
            return format

    return None


def parse_file(filename):
    with open(filename, 'rb') as f:
        image = f.read()

    base, ext = os.path.splitext(filename)
    format = detect_file_format(image, ext)
    if not format:
        raise zx.Error('Cannot determine the format of file %r.' % filename)

    return format().parse(image)


class TapePlayer(object):
    def __init__(self):
        self._is_paused = False
        self._pulses = []
        self._tick = 0
        self._level = False
        self._pulse = 0
        self._ticks_per_frame = 69888  # TODO

    def load_parsed_file(self, file):
        self._pulses = file.get_pulses()
        self._level = False
        self._is_paused = True
        print('Press F6 to unpause the tape.')

    def load_tape(self):
        filename = input('Tape file: ')
        self.load_parsed_file(parse_file(filename))

    def pause_or_unpause(self):
        self._is_paused = not self._is_paused
        print('Tape is %s.' % ('paused' if self._is_paused else 'unpaused'))

    def get_level_at_frame_tick(self, tick):
        assert self._tick <= tick

        while self._tick < tick:
            if self._is_paused:
                self._tick = tick
                continue

            # See if we already have a non-zero-length pulse.
            if self._pulse:
                ticks_to_skip = min(self._pulse, tick - self._tick)
                self._pulse -= ticks_to_skip
                self._tick += ticks_to_skip
                continue

            # Get subsequent pulse, if any.
            got_new_pulse = False
            for level, pulse in self._pulses:
                got_new_pulse = True
                self._level = level
                self._pulse = pulse
                # print(self._pulse)
                break

            # Do nothing, if there are no more pulses available.
            if not got_new_pulse:
                self._level = False
                self._tick = tick

        return self._level

    def skip_rest_of_frame(self):
        if self._tick < self._ticks_per_frame:
            self.get_level_at_frame_tick(self._ticks_per_frame)

        assert self._tick >= self._ticks_per_frame
        self._tick -= self._ticks_per_frame


class emulator(Gtk.Window):
    def __init__(self, speed_factor=1.0):
        super(emulator, self).__init__()

        self._speed_factor = speed_factor

        self.frame_width = 48 + 256 + 48
        self.frame_height = 48 + 192 + 40

        self.done = False
        self.is_paused = False

        self.scale = 2

        self.area = Gtk.DrawingArea()
        self.area.connect("draw", self.on_draw_area)
        self.add(self.area)

        self.set_title("ZX Spectrum Emulator")
        self.resize(self.frame_width * self.scale,
                    self.frame_height * self.scale)
        self.set_position(Gtk.WindowPosition.CENTER)
        self.connect("delete-event", self.on_done)

        # Don't even show the window on full throttle.
        if self._speed_factor is not None:
            self.show_all()

        self.frame_size = self.frame_width * self.frame_height
        self.frame = cairo.ImageSurface(cairo.FORMAT_RGB24,
                                        self.frame_width, self.frame_height)
        self.frame_data = self.frame.get_data()

        self.pattern = cairo.SurfacePattern(self.frame)
        self.pattern.set_filter(cairo.FILTER_NEAREST)

        self.emulator = zx.Spectrum48()
        self.processor_state = self.emulator  # TODO: Eliminate.

        self.keyboard_state = [0xff] * 8
        self.keys = {'RETURN': zx.KEYS_INFO['ENTER'],
                     'SHIFT_L': zx.KEYS_INFO['CAPS SHIFT'],
                     'SHIFT_R': zx.KEYS_INFO['SYMBOL SHIFT'],
                     'SPACE': zx.KEYS_INFO['BREAK SPACE']}
        for id in ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
                   'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
                   'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
                   'U', 'V', 'W', 'X', 'Y', 'Z']:
            self.keys[id] = zx.KEYS_INFO[id]
        self.emulator.set_on_input_callback(self.on_input)
        self.connect("key-press-event", self.on_key_press)
        self.connect("key-release-event", self.on_key_release)

        self.tape_player = TapePlayer()

    def on_done(self, widget, context):
        self.done = True

    def on_draw_area(self, widget, context):
        context.save()
        context.scale(self.scale, self.scale)
        context.set_source(self.pattern)
        context.paint()
        context.restore()

    def handle_spectrum_key(self, key_info, pressed):
        if not key_info:
            return

        # print(key_info['id'])
        addr_line = key_info['address_line']
        mask = 1 << key_info['port_bit']

        if pressed:
            self.keyboard_state[addr_line - 8] &= mask ^ 0xff
        else:
            self.keyboard_state[addr_line - 8] |= mask

    def show_help(self):
        KEYS = [
            ('F1', 'Show help.'),
            ('F2', 'Save snapshot.'),
            ('F3', 'Load tape file.'),
            ('F6', 'Pause/unpause tape.'),
            ('F10', 'Quit.'),
            ('PAUSE', 'Pause/unpause emulation.'),
        ]

        for entry in KEYS:
            print('%7s  %s' % entry)

    def pause_or_unpause(self):
        self.is_paused = not self.is_paused
        print('%s.' % ('Paused' if self.is_paused else 'Unpaused'))

    def on_key_press(self, widget, event):
        key_id = Gdk.keyval_name(event.keyval).upper()
        # print(key_id)

        if key_id in ['ESCAPE', 'F10']:
            self.done = True
            return

        if key_id == 'F1':
            self.show_help()

        if key_id == 'F2':
            self.save_snapshot()

        if key_id == 'F3':
            self.tape_player.load_tape()

        if key_id == 'F6':
            self.tape_player.pause_or_unpause()

        if key_id == 'PAUSE':
            self.pause_or_unpause()

        self.handle_spectrum_key(self.keys.get(key_id, None), pressed=True)

    def on_key_release(self, widget, event):
        key_id = Gdk.keyval_name(event.keyval).upper()
        self.handle_spectrum_key(self.keys.get(key_id, None), pressed=False)

    def on_input(self, addr):
        # Scan keyboard.
        n = 0xbf
        addr ^= 0xffff
        if addr & (1 << 8):
            n &= self.keyboard_state[0]
        if addr & (1 << 9):
            n &= self.keyboard_state[1]
        if addr & (1 << 10):
            n &= self.keyboard_state[2]
        if addr & (1 << 11):
            n &= self.keyboard_state[3]
        if addr & (1 << 12):
            n &= self.keyboard_state[4]
        if addr & (1 << 13):
            n &= self.keyboard_state[5]
        if addr & (1 << 14):
            n &= self.keyboard_state[6]
        if addr & (1 << 15):
            n &= self.keyboard_state[7]

        # TODO: Use the tick when the ear value is sampled
        #       instead of the tick of the beginning of the input
        #       cycle.
        tick = self.emulator.get_ticks_since_int()
        if self.tape_player.get_level_at_frame_tick(tick):
            n |= 0x40

        return n

    def save_snapshot(self):
        filename = input('Snapshot filename: ')
        with open(filename, 'wb') as f:
            f.write(zx.Z80SnapshotsFormat().make(self.emulator))
        print('%s saved.' % filename)

    def _find_recording_info_chunk(self, recording):
        for chunk in recording['chunks']:
            if chunk['id'] == 'info':
                return chunk
        assert 0  # TODO

    def _save_crash_rzx(self, recording, state, chunk_i, frame_i):
        snapshot = zx.Z80SnapshotsFormat().make(state)

        crash_recording = {
            'id': 'input_recording',
            'chunks': [
                self._find_recording_info_chunk(recording),
                {
                    'id': 'snapshot',
                    'image': snapshot,
                },
                {
                    'id': 'port_samples',
                    'first_tick': 0,
                    'frames': recording['chunks'][chunk_i]['frames'][frame_i:],
                },
            ],
        }

        with open('__crash.z80', 'wb') as f:
            f.write(snapshot)

        with open('__crash.rzx', 'wb') as f:
            f.write(zx.make_rzx(crash_recording))

    def playback_input_recording(self, file):
        # Interrupts are supposed to be controlled by the
        # recording.
        machine_state = self.emulator  # TODO: Eliminate.
        machine_state.suppress_int()
        machine_state.allow_int_after_ei(True)
        # machine_state.enable_trace()

        assert isinstance(file, RZXFile)
        recording = file._recording

        creator_info = self._find_recording_info_chunk(recording)

        # SPIN v0.5 alters ROM to implement fast tape loading,
        # but that affects recorded RZX files.
        spin_v0p5_info = {'id': 'info',
                          'creator': b'SPIN 0.5            ',
                          'creator_major_version': 0,
                          'creator_minor_version': 5 }
        if creator_info == spin_v0p5_info:
            machine_state.set_memory_block(0x1f47, b'\xf5')

        # The bytes-saving ROM procedure needs special processing.
        machine_state.set_breakpoint(0x04d4)

        # Process chunks in order.
        frame_count = 0
        chunks = recording['chunks']
        for chunk_i, chunk in enumerate(chunks):
            if self.done:
                break

            if isinstance(chunk, MachineSnapshot):
                self.emulator.install_snapshot(chunk)
                continue

            if chunk['id'] != 'port_samples':
                continue

            first_tick = chunk['first_tick']
            machine_state.set_ticks_since_int(first_tick)

            for frame_i, frame in enumerate(chunk['frames']):
                if self.done:
                    break

                frame_state = machine_state.clone()

                # TODO: For debug purposes.
                '''
                frame_count += 1
                if frame_count == -12820:
                    self._save_crash_rzx(recording, frame_state, chunk_i, frame_i)
                    assert 0

                if frame_count == -65952 - 1000:
                    machine_state.enable_trace()
                '''

                num_of_fetches, samples = frame
                # print(num_of_fetches, samples)
                self.sample_i = 0
                def on_input(addr):
                    # print(machine_state.get_fetches_limit())
                    fetch = num_of_fetches - machine_state.get_fetches_limit()
                    # print('Input at fetch', fetch, 'of', num_of_fetches)

                    if self.sample_i >= len(samples):
                        raise zx.Error(
                            'Too few input samples at frame %d of %d. '
                            'Given %d, used %d.' % (
                                frame_count, len(chunk['frames']),
                                len(samples), self.sample_i),
                            id='too_few_input_samples')

                    n = samples[self.sample_i]
                    # TODO: print('read_port 0x%04x 0x%02x' % (addr, n), flush=True)
                    self.sample_i += 1
                    return n

                self.emulator.set_on_input_callback(on_input)

                # TODO: print(num_of_fetches, samples, flush=True)
                machine_state.set_fetches_limit(num_of_fetches)

                while not self.done:
                    while Gtk.events_pending():
                        Gtk.main_iteration()

                    events = self.emulator.run() if not self.is_paused else 0
                    # TODO: print(events)

                    if events & machine_state._BREAKPOINT_HIT:
                        # SPIN v0.5 skips executing instructions
                        # of the bytes-saving ROM procedure in
                        # fast save mode.
                        if (creator_info == spin_v0p5_info and
                                machine_state.get_pc() == 0x04d4):
                            sp = machine_state.get_sp()
                            ret_addr = machine_state.read16(sp)
                            machine_state.set_sp(sp + 2)
                            machine_state.set_pc(ret_addr)

                    if (events & machine_state._END_OF_FRAME and
                            self._speed_factor is not None):
                        self.emulator.render_screen()
                        self.frame_data[:] = self.emulator.get_frame_pixels()
                        self.area.queue_draw()
                        # print(self.processor_state.get_bc())
                        time.sleep((1 / 50) * self._speed_factor)

                    if events & machine_state._FETCHES_LIMIT_HIT:
                        # Some simulators, e.g., SPIN, may store an interrupt
                        # point in the middle of a IX- or IY-prefixed
                        # instruction, so we continue until such
                        # instruction, if any, is completed.
                        if machine_state.get_iregp_kind() != 'hl':
                            machine_state.set_fetches_limit(1)
                            continue

                        # SPIN doesn't update the fetch counter if the last
                        # instruction in frame is IN.
                        if (creator_info == spin_v0p5_info and
                                self.sample_i < len(samples)):
                            machine_state.set_fetches_limit(1)
                            continue

                        self.emulator.on_handle_active_int()
                        break

                if self.sample_i != len(samples):
                    raise zx.Error(
                        'Too many input samples at frame %d of %d. '
                        'Given %d, used %d.' % (
                            frame_count, len(chunk['frames']),
                            len(samples), self.sample_i),
                        id='too_many_input_samples')


    def main(self):
        # self.emulator.enable_trace()

        while not self.done:
            while Gtk.events_pending():
                Gtk.main_iteration()

            events = self.emulator.run() if not self.is_paused else 0
            # TODO: print(events)

            if events & self.emulator._FETCHES_LIMIT_HIT:
                set_fetches_limit = True

            if events & self.emulator._END_OF_FRAME:
                self.emulator.render_screen()
                self.frame_data[:] = self.emulator.get_frame_pixels()
                self.area.queue_draw()
                self.tape_player.skip_rest_of_frame()
                time.sleep((1 / 50) * self._speed_factor)

    def run_file(self, filename):
        file = parse_file(filename)

        if isinstance(file, MachineSnapshot):
            self.emulator.install_snapshot(file)
            self.main()
        elif isinstance(file, RZXFile):
            self.playback_input_recording(file)
        elif isinstance(file, zx.SoundFile):
            self.tape_player.load_parsed_file(file)
            self.main()
        else:
            raise zx.Error("Don't know how to run file %r." % filename)


def handle_extra_arguments(args):
    if args:
        raise zx.Error('Extra argument %r.' % args[0])


def run(args):
    if not args:
        app = emulator()
        app.main()
    else:
        filename = args.pop(0)
        handle_extra_arguments(args)

        app = emulator()
        app.run_file(filename)


def dump(args):
    if not args:
        raise zx.Error('The file to dump is not specified.')

    filename = args.pop(0)
    handle_extra_arguments(args)

    file = parse_file(filename)
    print(file)


def looks_like_filename(s):
    return '.' in s


def usage():
    print('Usage:')
    print('  zx [run] [<filename>]')
    print('  zx dump <filename>')
    print('  zx help')
    sys.exit()


def test_file(filename):
    print('%r' % filename)
    def move(dest_dir):
        os.makedirs(dest_dir, exist_ok=True)

        # Make sure the destination filename is unique.
        dest_filename = filename
        while True:
            dest_path = os.path.join(dest_dir, dest_filename)
            if not os.path.exists(dest_path):
                break

            dest_filename, ext = os.path.splitext(dest_filename)
            dest_filename = dest_filename.rsplit('--', maxsplit=1)
            if len(dest_filename) == 1:
                dest_filename = dest_filename[0] + '--2'
            else:
                dest_filename = (dest_filename[0] + '--' +
                                     str(int(dest_filename[1]) + 1))

            dest_filename = dest_filename + ext


        os.rename(filename, dest_path)
        print('%r moved to %r' % (filename, dest_dir))

    app = emulator(speed_factor=None)
    try:
        app.run_file(filename)
        if app.done:
            return False
        move('passed')
    except zx.Error as e:
        move(e.id)

    app.destroy()

    return True


def test(args):
    for filename in args:
        if not test_file(filename):
            break


def fastforward(args):
    for filename in args:
        app = emulator(speed_factor=0)
        app.run_file(filename)
        if app.done:
            break

        app.destroy()


def convert_file(src_filename, dest_filename):
    src = parse_file(src_filename)
    src_format = src.get_format()
    # print(src, '->', dest_filename)

    _, dest_ext = os.path.splitext(dest_filename)
    dest_format = detect_file_format(image=None, filename_extension=dest_ext)
    if not dest_format:
        raise zx.Error('Cannot determine the format of file %r.' % (
                           dest_filename))

    if issubclass(src_format, zx.SoundFileFormat):
        if issubclass(dest_format, zx.SoundFileFormat):
            dest_format().save_from_pulses(dest_filename, src.get_pulses())
        else:
            raise zx.Error("Don't know how to convert from %s to %s files." % (
                               src_format().get_name(),
                               dest_format().get_name()))
    else:
        raise zx.Error("Don't know how to convert from %s files." % (
                           src_format().get_name()))


def convert(args):
    if not args:
        raise zx.Error('The file to convert from is not specified.')
    src_filename = args.pop(0)

    if not args:
        raise zx.Error('The file to convert to is not specified.')
    dest_filename = args.pop(0)

    handle_extra_arguments(args)

    convert_file(src_filename, dest_filename)


def handle_command_line(args):
    # Guess the command by the arguments.
    if (not args or
        len(args) == 1 and looks_like_filename(args[0])):
        run(args)
        return

    if (len(args) == 2 and looks_like_filename(args[0]) and
        looks_like_filename(args[1])):
        convert(args)
        return

    # Handle an explicitly specified command.
    command = args[0]
    if command in ['help', '-help', '--help',
                   '-h', '-?',
                   '/h', '/help']:
        usage()
        return

    COMMANDS = {
        'convert': convert,
        'dump': dump,
        'run': run,

        # TODO: Hidden commands for internal use.
        '__test': test,
        '__ff': fastforward,
    }

    if command not in COMMANDS:
        raise zx.Error('Unknown command %r.' % command)

    COMMANDS[command](args[1:])


def main():
    try:
        handle_command_line(sys.argv[1:])
    except zx.Error as e:
        print('zx: %s' % e.args)


if __name__ == "__main__":
    # import cProfile
    # cProfile.run('main()')
    main()



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


def get_standard_pilot_pulses(is_header):
    pulse = 2168
    duration = 8063 if is_header else 3223
    for _ in range(duration):
        yield pulse


def get_standard_sync_pulses():
    yield 667
    yield 735


def _get_data_bits(data):
    for byte in data:
        for i in range(8):
            yield (byte & (1 << (7 - i))) != 0


def get_standard_data_pulses(data):
    for bit in _get_data_bits(data):
        pulse = 1710 if bit else 855
        yield pulse
        yield pulse


def get_standard_block_pulses(data):
    # Generate pilot tone.
    is_header = data[0] < 128
    for pulse in get_standard_pilot_pulses(is_header):
        yield pulse

    # Sync pulses.
    for pulse in get_standard_sync_pulses():
        yield pulse

    # Data.
    for pulse in get_standard_data_pulses(data):
        yield pulse



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


import gc, unittest


class test_create(unittest.TestCase):
    def runTest(self):
        import zx
        mach = zx.Spectrum48()
        del mach


class test_derive(unittest.TestCase):
    def runTest(self):
        import zx
        class speccy(zx.Spectrum48):
            pass
        mach = speccy()
        del mach


class test_get_memory(unittest.TestCase):
    def runTest(self):
        import zx
        mach = zx.Spectrum48()
        mem = mach.get_memory()
        assert len(mem) == 0x10000
        assert mem[0] == 0xf3, '0x%x' % mem[0]
        mem[0] += 1
        assert mem[0] == 0xf4, '0x%x' % mem[0]


class test_render_frame(unittest.TestCase):
    def runTest(self):
        import zx
        mach = zx.Spectrum48()
        data = mach.render_screen()
        assert len(data) == 49280


class test_get_frame_pixels(unittest.TestCase):
    def runTest(self):
        import zx
        mach = zx.Spectrum48()
        mach.render_screen()
        pixels = mach.get_frame_pixels()
        assert len(pixels) == 394240


class test_execute_frame(unittest.TestCase):
    def runTest(self):
        import zx
        mach = zx.Spectrum48()

        mem = mach.get_memory()
        mem[0:2] = bytearray([0x18, 0x100 - 2])  # jr $

        mach.run()


class test_keyboard(unittest.TestCase):
    def runTest(self):
        import zx
        info = zx.KEYS_INFO
        assert info['3']['number'] == 2
        assert info['N']['halfrow_number'] == 7
        assert info['H']['pos_in_halfrow'] == 0
        assert info['C']['is_leftside'] == True
        assert info['E']['is_rightside'] == False
        assert info['R']['address_line'] == 10
        assert info['5']['port_bit'] == 4


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


from ._binary import BinaryParser, BinaryWriter
import zx


def parse_creator_info_block(image):
    parser = BinaryParser(image)
    chunk = parser.parse([('creator', '20s'),
                          ('creator_major_version', '<H'),
                          ('creator_minor_version', '<H')])
    chunk['id'] = 'info'
    return chunk


def parse_snapshot_block(image):
    parser = BinaryParser(image)
    header = parser.parse([('flags', '<L'),
                           ('filename_extension', '4s'),
                           ('uncompressed_length', '<L')])

    flags = header['flags']
    descriptor = bool(flags & 0x1)
    compressed = bool(flags & 0x2)

    # TODO: Support snapshot descriptors.
    if descriptor:
        raise zx.Error('RZX snapshot descriptors are not supported yet.',
                       id='rzx_snapshot_descriptor')

    snapshot_image = parser.extract_rest()
    if compressed:
        import zlib
        snapshot_image = zlib.decompress(snapshot_image)

    # TODO: Support other snapshot formats.
    filename_extension = header['filename_extension']
    if filename_extension not in [b'z80\x00', b'Z80\x00']:
        raise zx.Error('Unknown RZX snapshot format %r.' % filename_extension,
                       id='unknown_rzx_snapshot_format')

    format = zx.Z80SnapshotsFormat()
    return format.parse(snapshot_image)


def parse_input_recording_block(image):
    parser = BinaryParser(image)
    header = parser.parse([('num_of_frames', '<L'),
                           ('reserved', 'B'),
                           ('first_tick', '<L'),
                           ('flags', '<L')])

    flags = header['flags']
    protected = bool(flags & 0x1)
    compressed = bool(flags & 0x2)

    assert not protected  # TODO: Support protected samples.

    recording_image = parser.extract_rest()
    if compressed:
        import zlib
        recording_image = zlib.decompress(recording_image)

    recording_parser = BinaryParser(recording_image)

    NUM_OF_SAMPLES_IN_REPEATED_FRAME = 0xffff
    frames = []
    while not recording_parser.is_eof():
        recording_header = recording_parser.parse([
            ('num_of_fetches', '<H'),
            ('num_of_port_samples', '<H')])

        num_of_samples = recording_header['num_of_port_samples']
        num_of_fetches = recording_header['num_of_fetches']

        if num_of_samples == NUM_OF_SAMPLES_IN_REPEATED_FRAME:
            # Note that only the input samples are repeated; the
            # number of fetches is as specified in the new frame.
            assert frames  # TODO
            prev_samples = frames[-1][1]
            frame = (num_of_fetches, prev_samples)
        else:
            samples = recording_parser.extract_block(num_of_samples)
            samples = bytes(samples)
            frame = (num_of_fetches, samples)

        # Ignore empty frames.
        if frame[0] != 0:
            frames.append(frame)

    return {'id': 'port_samples',
            'first_tick': header['first_tick'],
            'frames': frames}


RZX_BLOCK_ID_CREATOR_INFO = 0x10
RZX_BLOCK_ID_SNAPSHOT = 0x30
RZX_BLOCK_ID_INPUT_RECORDING = 0x80


RZX_BLOCK_PARSERS = {
    RZX_BLOCK_ID_CREATOR_INFO: parse_creator_info_block,
    RZX_BLOCK_ID_SNAPSHOT: parse_snapshot_block,
    RZX_BLOCK_ID_INPUT_RECORDING: parse_input_recording_block,
}


def parse_block(parser):
    # Parse block header.
    block = parser.parse([('id', 'B'),
                          ('length', '<L')])

    # Extract block payload image.
    block_length = block['length']
    if block_length < 5:
        raise ZXError('RZX block length is too small: %d' % block_length)
    payload_size = block_length - 5
    payload_image = parser.extract_block(payload_size)

    # Parse payload image.
    block_id = block['id']

    # TODO: Handle unknown blocks.
    if block_id not in RZX_BLOCK_PARSERS:
        raise zx.Error('Unsupported RZX block %r.' % block_id,
                       id='unsupported_rzx_block')

    return RZX_BLOCK_PARSERS[block_id](payload_image)


def parse_rzx(image):
    parser = BinaryParser(image)

    # Unpack header.
    header = parser.parse([('signature', '4s'),
                           ('major_revision', 'B'),
                           ('minor_revision', 'B'),
                           ('flags', '<L')])
    rzx_signature = b'RZX!'
    if header['signature'] != rzx_signature:
        raise zx.Error('Bad RZX file signature %r; expected %r.' % (
                          header['signature'], rzx_signature))

    # Unpack blocks.
    chunks = []
    while not parser.is_eof():
        chunks.append(parse_block(parser))

    return {'id': 'input_recording', 'chunks': chunks}


def make_rzx(recording):
    assert recording['id'] == 'input_recording'

    writer = BinaryWriter()
    signature = b'RZX!'
    major_revision = b'\x00'
    minor_revision = b'\x0c'
    flags = b'\x00\x00\x00\x00'
    writer.write_block(signature + major_revision + minor_revision + flags)

    for chunk in recording['chunks']:
        chunk_writer = BinaryWriter()
        id = chunk['id']
        if id == 'info':
            chunk_id = RZX_BLOCK_ID_CREATOR_INFO
            chunk_writer.write(['20s:creator',
                                '<H:creator_major_version',
                                '<H:creator_minor_version'], **chunk)
        elif id == 'snapshot':
            chunk_id = RZX_BLOCK_ID_SNAPSHOT

            image = chunk['image']
            chunk_writer.write(['<L:flags', '4s:filename_extension',
                                '<L:uncompressed_length'],
                                flags=0,  # Non-descriptor. Not compressed.
                                filename_extension=b'Z80\x00',
                                uncompressed_length=len(image))
            chunk_writer.write_block(image)
        elif id == 'port_samples':
            chunk_id = RZX_BLOCK_ID_INPUT_RECORDING

            frames = chunk['frames']
            chunk_writer.write(['<L:num_of_frames', 'B:reserved',
                                '<L:first_tick', '<L:flags'],
                               num_of_frames=len(frames), reserved=0,
                               first_tick=0,  # TODO
                               flags=0,  # Not protected. Not compressed.
                              )

            for num_of_fetches, samples in frames:
                chunk_writer.write(['<H:num_of_fetches',
                                    '<H:num_of_port_samples'],
                                   num_of_fetches=num_of_fetches,
                                   num_of_port_samples=len(samples))
                chunk_writer.write_block(samples)
        else:
            assert 0, (id, list(chunk))  # TODO

        writer.write(['B:id', '<L:size'],
                      id=chunk_id, size=len(chunk_writer.get_image()) + 4 + 1)
        writer.write_block(chunk_writer.get_image())

    return writer.get_image()



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


from ._binary import BinaryParser
from ._tape import *
import zx


class TZXFile(zx.SoundFile):
    _TICKS_FREQ = 3500000

    def __init__(self, fields):
        zx.SoundFile.__init__(self, TZXFileFormat, fields)

    def get_pulses(self):
        level = False
        for block in self['blocks']:
            id = block['id']
            if id == '10 (Standard Speed Data Block)':
                # The block itself.
                data = block['data']
                for pulse in get_standard_block_pulses(data):
                    yield (level, pulse)
                    level = not level

                # Pause.
                pause_duration = block['pause_after_block_in_ms']

                # Pause blocks of zero duration shall be ignored,
                # and then the output level remains the same.
                if pause_duration == 0:
                    continue

                # At the end of the non-zero-duration pause the
                # output level shall be low.
                ''' TODO: Despite the specification, this cases
                          tape loading errors.
                if level:
                    # Give the high pulse 1ms of time and drop it.
                    yield (level, self._TICKS_FREQ / 1000)
                    pause_duration -= 1
                    level = not level

                assert not level
                '''

                if pause_duration:
                    yield (level, pause_duration * self._TICKS_FREQ / 1000)
            elif id == '30 (Text Description)':
                print(block['text'])
            else:
                assert 0, block  # TODO


class TZXFileFormat(zx.SoundFileFormat):
    _NAME = 'TZX'

    def _parse_standard_speed_data_block(self, parser):
        block = parser.parse([('pause_after_block_in_ms', '<H'),
                              ('data_size', '<H')])
        block.update({'id': '10 (Standard Speed Data Block)',
                      'data': parser.extract_block(block['data_size'])})
        del block['data_size']
        return block

    def _parse_text_description(self, parser):
        size = parser.parse_field('B', 'text_size')
        text = parser.extract_block(size)
        return {'id': '30 (Text Description)',
                'text': text}

    _BLOCK_PARSERS = {
        0x10: _parse_standard_speed_data_block,
        0x30: _parse_text_description,
    }

    def _parse_block(self, parser):
        block_id = parser.parse_field('B', 'block_id')
        if block_id not in self._BLOCK_PARSERS:
            raise zx.Error('Unsupported TZX block id %x.' % block_id)

        return self._BLOCK_PARSERS[block_id](self, parser)

    def parse(self, image):
        parser = BinaryParser(image)

        # Parse header.
        header = parser.parse([('signature', '8s'),
                               ('major_revision', 'B'),
                               ('minor_revision', 'B')])
        tzx_signature = b'ZXTape!\x1a'
        if header['signature'] != tzx_signature:
            raise zx.Error('Bad TZX file signature %r; expected %r.' % (
                              header['signature'], tzx_signature))

        # Parse blocks.
        blocks = []
        while not parser.is_eof():
            blocks.append(self._parse_block(parser))

        return TZXFile({**header, 'blocks': blocks})



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


KEYS_INFO = dict()

# Generate layout-related information.
layout = ['1', '2', '3', '4', '5', '6', '7', '8', '9', '0',
          'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I', 'O', 'P',
          'A', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', 'ENTER',
          'CAPS SHIFT', 'Z', 'X', 'C', 'V', 'B', 'N', 'M', 'SYMBOL SHIFT',
          'BREAK SPACE']
for n, id in enumerate(layout):
    key = KEYS_INFO.setdefault(id, dict())
    key['id'] = id
    key['number'] = n  # Left to right, then top to bottom.
    key['halfrow_number'] = n // 5
    key['pos_in_halfrow'] = n % 5
    key['is_leftside'] = key['halfrow_number'] % 2 == 0
    key['is_rightside'] = not key['is_leftside']

# Compute port address lines and bit positions.
for id, key in KEYS_INFO.items():
    if key['is_leftside']:
        key['address_line'] = 11 - key['halfrow_number'] // 2
        key['port_bit'] = key['pos_in_halfrow']
    else:
        key['address_line'] = key['halfrow_number'] // 2 + 12
        key['port_bit'] = 4 - key['pos_in_halfrow']



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


import struct, wave
import zx


class WAVFileFormat(zx.SoundFileFormat):
    _NAME = 'WAV'
    _TICKS_FREQ = 3500000

    def save_from_pulses(self, filename, pulses):
        with wave.open(filename, 'wb') as f:
            f.setnchannels(1)
            f.setsampwidth(2)

            frame_rate = 44100
            f.setframerate(frame_rate)

            amplitude = 32767
            for level, duration in pulses:
                duration = int(duration * frame_rate / self._TICKS_FREQ)
                sample = -amplitude if level else amplitude
                frame = struct.pack('<h', sample)
                f.writeframes(frame * duration)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import distutils.sysconfig, os
import inspect
# from distutils.core import Extension
from setuptools import Extension, setup


ZX_MAJOR_VERSION = 0
ZX_MINOR_VERSION = 4
ZX_PATCH_VERSION = 0


here = os.path.abspath(os.path.dirname(inspect.getsource(lambda:0)))

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


# Work around the problem with the warning about '-Wstrict-prototypes'.
# https://bugs.python.org/issue1222585
config_vars = distutils.sysconfig.get_config_vars()
opt_to_remove = '-Wstrict-prototypes'
for var in ['OPT']:
    if var in config_vars:
        opts = config_vars[var].split()
        if opt_to_remove in opts:
            opts.remove(opt_to_remove)
    config_vars[var] = ' '.join(opts)


zx_emulator_module = Extension(
    name='zx._emulator',
    define_macros=[('ZX_MAJOR_VERSION', '%d' % ZX_MAJOR_VERSION),
                   ('ZX_MINOR_VERSION', '%d' % ZX_MINOR_VERSION),
                   ('ZX_PATCH_VERSION', '%d' % ZX_PATCH_VERSION)],
    extra_compile_args=['-std=c++11', '-Wall', '-fno-exceptions', '-fno-rtti',
                        '-O3',
                        '-UNDEBUG',  # TODO
                       ],
    sources=['zx.cpp', 'zx/_emulatormodule.cpp'],
    language='c++')


# TODO: Update the URL once we have a published documentation.
# TODO: Do we have a name for the emulator?
setup(name='zx',
      version='%d.%d.%d' % (ZX_MAJOR_VERSION, ZX_MINOR_VERSION,
                            ZX_PATCH_VERSION),
      description='ZX Spectrum Emulator',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Ivan Kosarev',
      author_email='ivan@kosarev.info',
      url='https://github.com/kosarev/zx/',
      ext_modules=[zx_emulator_module],
      packages=['zx'],
      install_requires=[
          'pycairo',
          'pygobject',
      ],
      entry_points={
          'console_scripts': [
              'zx = zx:main',
          ],
      },
      classifiers=[
          'Development Status :: 2 - Pre-Alpha',
          'Environment :: X11 Applications :: GTK',
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: End Users/Desktop',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Programming Language :: C++',
          # TODO: Are we going to support Python 2?
          # TODO: Specific versions?
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: Implementation :: CPython',
          'Topic :: Games/Entertainment',
          'Topic :: Software Development',
          'Topic :: Software Development :: Libraries',
          'Topic :: System :: Emulators',
      ],
      license='MIT',
      # TODO: Respect other parameters.
      )



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


import struct, zx
from ._emulator import Spectrum48Base


class StateImage(object):
    def __init__(self, image):
        self._fields = {}
        self._image = image

    def define_fields(self, fields):
        self._fields.update(fields)

    def get(self, field_id):
        offset, format = self._fields[field_id]
        size = struct.calcsize(format)
        return struct.unpack(format, self._image[offset:offset + size])[0]

    def set(self, field_id, field_value):
        offset, format = self._fields[field_id]
        size = struct.calcsize(format)
        self._image[offset:offset + size] = struct.pack(format, field_value)


class ProcessorState(StateImage):
    _PROCESSOR_FIELDS = {
        'bc': (0, '<H'),
        'de': (2, '<H'),
        'hl': (4, '<H'),
        'af': (6, '<H'), 'f': (6, 'B'), 'a': (7, 'B'),
        'ix': (8, '<H'),
        'iy': (10, '<H'),

        'alt_bc': (12, '<H'),
        'alt_de': (14, '<H'),
        'alt_hl': (16, '<H'),
        'alt_af': (18, '<H'), 'alt_f': (18, 'B'), 'alt_a': (19, 'B'),

        'pc': (20, '<H'),
        'sp': (22, '<H'),
        'ir': (24, '<H'), 'r': (24, 'B'), 'i': (25, 'B'),
        'wz': (26, '<H'),

        'iff1': (28, 'B'),
        'iff2': (29, 'B'),
        'int_mode': (30, 'B'),
        'index_rp_kind': (31, 'B'),
    }

    def __init__(self, image):
        StateImage.__init__(self, image)
        self.define_fields(self._PROCESSOR_FIELDS)

    def get_bc(self):
        return self.get('bc')

    def get_de(self):
        return self.get('de')

    def get_hl(self):
        return self.get('hl')

    def get_a(self):
        return self.get('a')

    def get_f(self):
        return self.get('f')

    def get_ix(self):
        return self.get('ix')

    def get_iy(self):
        return self.get('iy')

    def get_alt_bc(self):
        return self.get('alt_bc')

    def get_alt_de(self):
        return self.get('alt_de')

    def get_alt_hl(self):
        return self.get('alt_hl')

    def get_alt_a(self):
        return self.get('alt_a')

    def get_alt_f(self):
        return self.get('alt_f')

    def get_alt_af(self):
        return self.get('alt_af')

    def get_pc(self):
        return self.get('pc')

    def set_pc(self, pc):
        self.set('pc', pc)

    def get_sp(self):
        return self.get('sp')

    def set_sp(self, sp):
        self.set('sp', sp)

    def get_i(self):
        return self.get('i')

    def get_r_reg(self):
        return self.get('r')

    def get_iff1(self):
        return self.get('iff1')

    def get_iff2(self):
        return self.get('iff2')

    def get_int_mode(self):
        return self.get('int_mode')

    def get_iregp_kind(self):
        n = self.get('index_rp_kind')
        return {0: 'hl', 1: 'ix', 2: 'iy'}[n]

    def install_snapshot(self, snapshot):
        assert isinstance(snapshot, zx.ProcessorSnapshot)
        for field, value in snapshot.items():
            if field != 'id':
                self.set(field, value)


class MemoryState(object):
    def __init__(self, image):
        self._memory_image = image

    def get_memory_block(self, addr, size):
        return self._memory_image[addr:addr + size]

    def set_memory_block(self, addr, block):
        self._memory_image[addr:addr + len(block)] = block

    def set_memory_blocks(self, blocks):
        for addr, block in blocks:
            self.set_memory_block(addr, block)

    def read8(self, addr):
        return self._memory_image[addr]

    def read16(self, addr):
        return zx.make16(hi=self.read8(addr + 1), lo=self.read8(addr))


class MachineState(ProcessorState, MemoryState):
    _MACHINE_FIELDS = {
        'ticks_since_int': (32, '<L'),
        'fetches_to_stop': (36, '<L'),
        'int_suppressed':  (40, 'B'),
        'int_after_ei_allowed': (41, 'B'),
        'border_color': (42, 'B'),
        'trace_enabled': (43, 'B'),
    }

    def __init__(self, machine_image, memory_image):
        ProcessorState.__init__(self, machine_image)
        MemoryState.__init__(self, memory_image)
        self.define_fields(self._MACHINE_FIELDS)

    def clone(self):
        return MachineState(self._image[:], self._memory_image[:])

    def is_suppressed_int(self):
        return bool(self.get('int_suppressed'))

    def suppress_int(self, suppress=True):
        self.set('int_suppressed', int(suppress))

    def allow_int_after_ei(self, allow=True):
        self.set('int_after_ei_allowed', int(allow))

    def get_fetches_limit(self):
        return self.get('fetches_to_stop')

    def set_fetches_limit(self, fetches_to_stop):
        self.set('fetches_to_stop', fetches_to_stop)

    def get_ticks_since_int(self):
        return self.get('ticks_since_int')

    def set_ticks_since_int(self, ticks):
        self.set('ticks_since_int', ticks)

    def get_border_color(self):
        return self.get('border_color')

    def set_border_color(self, color):
        self.set('border_color', color)

    def enable_trace(self, enable=True):
        self.set('trace_enabled', int(enable))

    def install_snapshot(self, snapshot):
        assert isinstance(snapshot, zx.MachineSnapshot)
        for field, value in snapshot.get_unified_snapshot().items():
            if field == 'processor_snapshot':
                ProcessorState.install_snapshot(self, value)
            elif field == 'memory_blocks':
                self.set_memory_blocks(value)
            else:
                self.set(field, value)


class Spectrum48(Spectrum48Base, MachineState):
    # Events.
    _NO_EVENTS         = 0
    _END_OF_FRAME      = 1 << 1
    _FETCHES_LIMIT_HIT = 1 << 3
    _BREAKPOINT_HIT    = 1 << 4

    # Memory marks.
    _NO_MARKS        = 0
    _BREAKPOINT_MARK = 1 << 0

    def __init__(self):
        self.machine_kind = 'ZX Spectrum 48K'
        MachineState.__init__(self, self.get_state_image(), self.get_memory())

        # Install ROM.
        self.set_memory_block(0x0000, zx.get_rom_image(self.machine_kind))

    def set_breakpoint(self, pc):
        self.mark_addrs(pc, 1, self._BREAKPOINT_MARK)



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.

import zx


class Data(object):
    def __init__(self, fields):
        self._fields = fields

    def __contains__(self, id):
        return id in self._fields

    def __getitem__(self, id):
        return self._fields[id]

    def __repr__(self):
        return repr(self._fields)

    def __iter__(self):
        for id in self._fields:
            yield id

    def items(self):
        for field in self._fields.items():
            yield field


class File(Data):
    def __init__(self, format, fields):
        self._format = format
        Data.__init__(self, fields)

    def get_format(self):
        return self._format


class FileFormat(object):
    def get_name(self):
        return self._NAME


class SoundFile(File):
    pass


class SoundFileFormat(FileFormat):
    pass
